import urine


def assert_equal(val):
    assert urine.decode(urine.encode(val)) == val


def assert_binary_size(val, n):
    assert len(urine.encode(val)) == 3 + n


def assert_int(val):
    assert isinstance(urine.decode(urine.encode(val)), int)


def test_int8_and_uint8():
    values = [35, -66, -0x80, 220, 0xFF, 0]
    for val in values:
        assert_equal(val)
        assert_binary_size(val, 1)
        assert_int(val)


def test_int16_and_uint16():
    values = [462, -844, -0x8000, 0xFFFF]
    for val in values:
        assert_equal(val)
        assert_binary_size(val, 2)
        assert_int(val)


def test_int32_and_uint32():
    values = [830007, -66220, -0x80000000, 0xFFFFFFFF]
    for val in values:
        assert_equal(val)
        assert_binary_size(val, 4)
        assert_int(val)


def test_int64_and_uint64():
    values = [600380402146, -396395892396, -0x8000000000000000, 0xFFFFFFFFFFFFFFFF]
    for val in values:
        assert_equal(val)
        assert_binary_size(val, 8)
        assert_int(val)


def test_bignum():
    values = [
        0xABCDEF0123456789ABCDEF0123456,
        -0x4876495867459DCDE45884AECD24A3583ACE398693EEAAA93CDE24595AE
    ]

    for val in values:
        assert_equal(val)
        assert_int(val)



import pytest

import urine


class TestClassDependency(object):
    data = [1, 3, 'test', 'check out porcupine tree (awesome band)']


class TestClass(object):
    string = 'blob'
    dictionary = {3: '3', 22: b'somebyteshereandthere', 1: {}}
    byte = b'\x00\xef'
    num1 = 33
    num2 = 314.235
    dependency = TestClassDependency()


@pytest.fixture(scope='module')
def test_class_instance():
    return TestClass()


@pytest.fixture(scope='module')
def test_class_dependency():
    return TestClassDependency


@pytest.fixture
def blank_encoder():
    return urine.UrineEncoder()


@pytest.fixture(scope='module')
def blank_decoder():
    return urine.UrineDecoder()



class UrineExtension:
    """Base class that every urine extension implements
    """
    def encode(obj):
        """Encodes the extension object.

        Args:
            obj (object): Object to be encoded

        Returns:
            bytes, bytearray: Encoded object

        Raises:
            NotImplementedError: Function not implemented
        """
        raise NotImplementedError('Extension has no encode() function implemented')

    def decode(data):
        """Decodes the extension object.

        Args:
            data (bytearray): Data to be decoded

        Returns:
            object: Decoded object

        Raises:
            NotImplementedError: Function not implemented
        """
        raise NotImplementedError('Extension has no decode() function implemented')



import urine


def test_complex():
    values = [complex(23, 5), complex(0, 0), complex(0.3563, -33), complex(-24.4466, -22)]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, complex)



import urine


@urine.exclude('attr1', 'attr3')
class ExcludeTestClass(object):
    attr1 = 420
    attr2 = 1337
    attr3 = 'hi'
    attr4 = []


@urine.include('attr2', 'attr4')
class IncludeTestClass(object):
    attr1 = 420
    attr2 = 1337
    attr3 = 'hi'
    attr4 = []


def test_class(test_class_instance, test_class_dependency):
    enc = urine.encode(test_class_instance)
    dec = urine.decode(enc)

    assert type(dec).__name__ == type(test_class_instance).__name__

    for attr in dir(test_class_instance):
        if not callable(attr) and not (attr.startswith('__') and attr.endswith('__')):
            dec_attr = getattr(dec, attr)
            if attr == 'dependency':
                assert dec_attr.data == test_class_instance.dependency.data
            else:
                assert dec_attr == getattr(test_class_instance, attr)


def test_class_with_excluded_attributes():
    instance = ExcludeTestClass()
    enc = urine.encode(instance)
    dec = urine.decode(enc)

    assert not hasattr(dec, 'attr1') and not hasattr(dec, 'attr3')
    assert hasattr(dec, 'attr2') and hasattr(dec, 'attr4')


def test_class_with_included_attributes():
    instance = IncludeTestClass()
    enc = urine.encode(instance)
    dec = urine.decode(enc)

    assert not hasattr(dec, 'attr1') and not hasattr(dec, 'attr3')
    assert hasattr(dec, 'attr2') and hasattr(dec, 'attr4')



import urine


def test_tuple():
    values = [tuple(), (3, 6, 1, 9)]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, tuple)


def test_nested_tuples():
    values = [tuple(tuple(tuple())), (2, 4, tuple((7, 3, (8,))))]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, tuple)



import urine


def test_string():
    values = ['', 'Just A Fucking Test String', '異體字字典‚∂ƒ∆€€©¶[ø']
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, str)



import urine

values = [b'sdgvn4309', b'\x00', b'\xfe\x02\xc2\x24\xae']


def test_bytes():
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, bytes)


def test_bytearray():
    for val in values:
        val = bytearray(val)
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, bytearray)



__protocol_version__ = '1.0'


from urine.decoder import UrineDecoder
from urine.encoder import UrineEncoder, exclude, include
from urine.extensions import UrineExtension

_encoder = UrineEncoder()
_decoder = UrineDecoder()


def encode(obj):
    """Encodes a python object.

    Args:
        obj (object): Object to be encoded

    Returns:
        bytearray: Encoded object
    """
    return _encoder.encode(obj)


def decode(data):
    """Decodes an object that was encoded with urine.

    Args:
        data (bytes, bytearray): Data to be decoded

    Returns:
        object: Decoded object
    """
    return _decoder.decode(data)


def extend(obj_type, extension):
    """Adds an extension to the encoder and decoder

    Args:
        obj_type (type): Object type the extension will apply to
        extension (UrineExtension): Extension that inherits and implements UrineExtension
    """
    _encoder.extend(obj_type, extension)
    _decoder.extend(extension)



import urine


def test_set():
    values = [set(), {3, 6, 'uiiii'}]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, set)


def test_set_with_nested_frozensets():
    values = [{frozenset((frozenset(),))}, {3, 'fuck', frozenset((3, 5))}]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, set)


def test_frozenset():
    values = [frozenset(), frozenset((1, '4', 6.44))]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, frozenset)


def test_nested_frozensets():
    values = [frozenset((frozenset(),)), frozenset((frozenset(), 3, 5, frozenset((3, 5.33))))]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, frozenset)



import urine


def test_dict():
    values = [{}, {1: 4, 'loong': 33.3}]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, dict)


def test_nested_dicts():
    values = [{3: 'rrtt', 33.3: {3: {}}}, {3: {}}]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, dict)



import urine


def test_bool():
    values = [True, False]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, bool)



import urine


def test_float():
    values = [0.4564, 0.0, -33.1164, -0.0001, 345346.3344874457457547]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, float)



import datetime
import struct

import pytest

import urine


class DatetimeExtension(urine.UrineExtension):
    def encode(obj):
        return struct.pack('<HBBBBBI', *[
            obj.year,
            obj.month,
            obj.day,
            obj.hour,
            obj.minute,
            obj.second,
            obj.microsecond
        ])

    def decode(data):
        unpacked = struct.unpack('<HBBBBBI', data)
        return datetime.datetime(*unpacked)


class MissingEncodeExtension(urine.UrineExtension):
    pass


class MissingDecodeExtension(urine.UrineExtension):
    def encode(obj):
        return b"TEST DATA"


class u202104(urine.UrineExtension):
    pass


class F108909(urine.UrineExtension):
    pass


def test_extensions(blank_encoder, blank_decoder):
    blank_encoder.extend(datetime.datetime, DatetimeExtension)
    blank_decoder.extend(DatetimeExtension)

    dt = datetime.datetime.today()
    enc = blank_encoder.encode(dt)
    dec = blank_decoder.decode(enc)

    assert dec == dt
    assert isinstance(dec, datetime.datetime)


def test_global_extensions():
    urine.extend(datetime.datetime, DatetimeExtension)

    dt = datetime.datetime.today()
    enc = urine.encode(dt)
    dec = urine.decode(enc)

    assert dec == dt
    assert isinstance(dec, datetime.datetime)


def test_duplicate_extensions():
    urine.extend(datetime.datetime, DatetimeExtension)
    urine.extend(datetime.datetime, DatetimeExtension)


def test_extension_with_missing_encode_function(blank_encoder, blank_decoder):
    blank_encoder.extend(datetime.datetime, MissingEncodeExtension)
    blank_decoder.extend(MissingEncodeExtension)

    dt = datetime.datetime.today()

    pytest.raises(NotImplementedError, lambda: blank_encoder.encode(dt))


def test_extension_with_missing_decode_function(blank_encoder, blank_decoder):
    blank_encoder.extend(datetime.datetime, MissingDecodeExtension)
    blank_decoder.extend(MissingDecodeExtension)

    dt = datetime.datetime.today()
    enc = blank_encoder.encode(dt)

    pytest.raises(NotImplementedError, lambda: blank_decoder.decode(enc))


def test_extensions_with_same_crc_hash_value(blank_encoder, blank_decoder):
    blank_encoder.extend(int, u202104)
    blank_decoder.extend(u202104)

    pytest.raises(ValueError, lambda: blank_encoder.extend(float, F108909))
    pytest.raises(ValueError, lambda: blank_decoder.extend(F108909))



import urine


def test_list():
    values = [[], [1, 2, 3]]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, list)


def test_nested_lists():
    values = [[[[[]]]], [1, 2, [1, [2], 3, [4, 5, 6]]]]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, list)



data_types = [
    'extension',

    'bool',

    'int8',
    'int16',
    'int32',
    'int64',
    'uint8',
    'uint16',
    'uint32',
    'uint64',
    'bignum',

    'float',
    'complex',

    'bytes',
    'bytearray',

    'string',

    'list',
    'tuple',
    'set',
    'frozenset',
    'dict',

    'range',

    'class',
    'none',
]



import urine


def test_nested_types():
    data = [
        [33.4, {4: b'eerrr', 10: set(), 20: frozenset(['ss', b'gg'])}, {44.6, 3, complex(3, 22.2)}],
        ({}, {3: True}), None,
        [0x457348976498fea73647ad37463eda24dafe3174ecad828defacd7392dace2845cc382742aec373aecd3925eadc],
        None,
        False
    ]

    enc = urine.encode(data)
    dec = urine.decode(enc)

    assert dec == data



import struct
import zlib

from urine import __protocol_version__
from urine.extensions import UrineExtension
from urine.types import data_types

_exclude_identifier = '_urine_exclude'
_include_identifier = '_urine_include'


def include(*args):
    """Class decorator specifying the class attributes to be encoded
    """
    def include(cls):
        setattr(cls, _include_identifier, args)
        return cls
    return include


def exclude(*args):
    """Class decorator specifying the class attributes to be excluded from encoding
    """
    def exclude(cls):
        setattr(cls, _exclude_identifier, args)
        return cls
    return exclude


class UrineEncoder(object):
    _type_id_dict = {x: i for i, x in enumerate(data_types)}

    def __init__(self):
        self.buffer = bytearray()

        self._extension_hash_dict = {}
        self._type_extension_dict = {}

    def extend(self, obj_type, extension):
        """Adds an extension to the encoder.

        Args:
            obj_type (type): Object type the extension will apply to
            extension (UrineExtension): Extension that inherits and implements UrineExtension

        Raises:
            ValueError: Two different extensions generate the same CRC32 hash (very unlikely)
        """
        if issubclass(extension, UrineExtension):
            if (extension in self._extension_hash_dict) or (obj_type in self._type_extension_dict):
                return

            crc32 = zlib.crc32(extension.__name__.encode('utf-8'))

            if crc32 in self._extension_hash_dict.values():
                raise ValueError('Extension hash duplicate detected, change extension name: {}'.format(
                    extension.__name__
                ))

            self._extension_hash_dict[extension] = crc32
            self._type_extension_dict[obj_type] = extension

    def encode(self, obj):
        """Encodes a python object.

        Args:
            obj (object): Object to be encoded

        Returns:
            bytearray: Encoded object
        """
        self._write_protocol_version()
        self._encode(obj)

        temp = self.buffer
        self.buffer = bytearray()

        return temp

    def encode_extension(self, obj, extension, add_type=True):
        if add_type:
            self._write_type('extension')

        encoded = extension.encode(obj)

        self.encode_uint32(self._extension_hash_dict[extension], add_type=False)
        self.encode_uint32(len(encoded), add_type=False)
        self.buffer += encoded

    def encode_bool(self, val, add_type=True):
        if add_type:
            self._write_type('bool')

        self.buffer += struct.pack('?', val)

    def encode_int(self, val, add_type=True):
        if val >= 0:
            if val <= 0xff:
                self.encode_uint8(val, add_type=add_type)
            elif val <= 0xffff:
                self.encode_uint16(val, add_type=add_type)
            elif val <= 0xffffffff:
                self.encode_uint32(val, add_type=add_type)
            elif val <= 0xffffffffffffffff:
                self.encode_uint64(val, add_type=add_type)
            else:
                self.encode_bignum(val, add_type=add_type)
        else:
            abs_val = abs(val)

            if abs_val <= 0x80:
                self.encode_int8(val, add_type=add_type)
            elif abs_val <= 0x8000:
                self.encode_int16(val, add_type=add_type)
            elif abs_val <= 0x80000000:
                self.encode_int32(val, add_type=add_type)
            elif abs_val <= 0x8000000000000000:
                self.encode_int64(val, add_type=add_type)
            else:
                self.encode_bignum(val, add_type=add_type)

    def encode_int8(self, val, add_type=True):
        if add_type:
            self._write_type('int8')

        self.buffer += struct.pack('b', val)

    def encode_int16(self, val, add_type=True):
        if add_type:
            self._write_type('int16')

        self.buffer += struct.pack('<h', val)

    def encode_int32(self, val, add_type=True):
        if add_type:
            self._write_type('int32')

        self.buffer += struct.pack('<i', val)

    def encode_int64(self, val, add_type=True):
        if add_type:
            self._write_type('int64')

        self.buffer += struct.pack('<q', val)

    def encode_uint8(self, val, add_type=True):
        if add_type:
            self._write_type('uint8')

        self.buffer += struct.pack('B', val)

    def encode_uint16(self, val, add_type=True):
        if add_type:
            self._write_type('uint16')

        self.buffer += struct.pack('<H', val)

    def encode_uint32(self, val, add_type=True):
        if add_type:
            self._write_type('uint32')

        self.buffer += struct.pack('<I', val)

    def encode_uint64(self, val, add_type=True):
        if add_type:
            self._write_type('uint64')

        self.buffer += struct.pack('<Q', val)

    def encode_bignum(self, val, add_type=True):
        if add_type:
            self._write_type('bignum')

        self.encode_bool(val < 0, add_type=False)

        val = hex(abs(val))[2:]
        parts = [val[i:i + 16] for i in range(0, len(val), 16)]

        self.encode_uint32(len(parts), add_type=False)

        for part in parts:
            self.encode_uint64(int(part, 16), add_type=False)

    def encode_float(self, val, add_type=True):
        if add_type:
            self._write_type('float')

        self.buffer += struct.pack('<d', val)

    def encode_complex(self, val, add_type=True):
        if add_type:
            self._write_type('complex')

        self.buffer += struct.pack('<dd', val.real, val.imag)

    def encode_bytes(self, data, add_type=True):
        if add_type:
            self._write_type('bytes')

        self.encode_uint32(len(data), add_type=False)
        self.buffer += data

    def encode_bytearray(self, data, add_type=True):
        if add_type:
            self._write_type('bytearray')

        self.encode_uint32(len(data), add_type=False)
        self.buffer += data

    def encode_string(self, string, add_type=True):
        byte_str = string.encode('utf-8')
        if add_type:
            self._write_type('string')

        self.encode_uint32(len(byte_str), add_type=False)
        self.buffer += byte_str

    def encode_list(self, lst, add_type=True):
        if add_type:
            self._write_type('list')

        self.encode_uint32(len(lst), add_type=False)

        for obj in lst:
            self._encode(obj)

    def encode_tuple(self, tpl, add_type=True):
        if add_type:
            self._write_type('tuple')

        self.encode_uint32(len(tpl), add_type=False)

        for obj in tpl:
            self._encode(obj)

    def encode_set(self, st, add_type=True):
        if add_type:
            self._write_type('set')

        self.encode_uint32(len(st), add_type=False)

        for obj in st:
            self._encode(obj)

    def encode_frozenset(self, st, add_type=True):
        if add_type:
            self._write_type('frozenset')

        self.encode_uint32(len(st), add_type=False)

        for obj in st:
            self._encode(obj)

    def encode_dict(self, dct, add_type=True):
        if add_type:
            self._write_type('dict')

        self.encode_uint32(len(dct), add_type=False)

        for key, val in dct.items():
            self._encode(key)
            self._encode(val)

    def encode_range(self, rng, add_type=True):
        if add_type:
            self._write_type('range')

        self.encode_int(rng.start)
        self.encode_int(rng.stop)
        self.encode_int(rng.step)

    def encode_class(self, obj, add_type=True):
        inclusions = () if not hasattr(obj, _include_identifier) else getattr(obj, _include_identifier)
        exclusions = () if not hasattr(obj, _exclude_identifier) else getattr(obj, _exclude_identifier)

        if inclusions:
            attrs = {x: getattr(obj, x) for x in inclusions}
        else:
            attrs = {
                x: getattr(obj, x)
                for x in dir(obj)
                if all((
                    not callable(getattr(obj, x)),
                    not (x.startswith('__') and x.endswith('__')),
                    x not in exclusions,
                    x != _exclude_identifier
                ))
            }

        if add_type:
            self._write_type('class')

        self.encode_string(type(obj).__name__, add_type=False)
        self.encode_uint32(len(attrs), add_type=False)

        for attr_name, val in attrs.items():
            self.encode_string(attr_name, add_type=False)
            self._encode(val)

    def encode_none(self):
        self._write_type('none')

    def _encode(self, obj):
        obj_type = type(obj)

        if obj_type in self._type_extension_dict:
            self.encode_extension(obj, self._type_extension_dict[obj_type])

        elif obj_type is bool:
            self.encode_bool(obj)
        elif obj_type is int:
            self.encode_int(obj)
        elif obj_type is float:
            self.encode_float(obj)
        elif obj_type is complex:
            self.encode_complex(obj)
        elif obj_type is bytes:
            self.encode_bytes(obj)
        elif obj_type is bytearray:
            self.encode_bytearray(obj)
        elif obj_type is str:
            self.encode_string(obj)
        elif obj_type is list:
            self.encode_list(obj)
        elif obj_type is tuple:
            self.encode_tuple(obj)
        elif obj_type is set:
            self.encode_set(obj)
        elif obj_type is frozenset:
            self.encode_frozenset(obj)
        elif obj_type is dict:
            self.encode_dict(obj)
        elif obj_type is range:
            self.encode_range(obj)
        elif isinstance(obj, object):
            if obj is None:
                self.encode_none()
            else:
                self.encode_class(obj)

    def _write_protocol_version(self):
        major, minor = __protocol_version__.split('.')

        self.encode_uint8(int(major), add_type=False)
        self.encode_uint8(int(minor), add_type=False)

    def _write_type(self, type_str):
        type_id = self._type_id_dict[type_str]
        self.encode_uint8(type_id, add_type=False)



import setuptools

with open('README.md', 'r') as f:
    long_description = f.read()

tests_require = [
    'pytest>=7.0.0',
    'pytest-cov>=3.0.0',
]

lint_requires = [
    'flake8>=4.0.1',
    'isort>=5.10.1',
]

dev_requires = [
    *tests_require,
    *lint_requires,
    'tox>=3.25.0',
]


setuptools.setup(
    name='urine',
    version='0.1.0',
    author='Christian Klein',
    author_email='klein_chris@outlook.com',
    description='Simple and secure binary serialization for Python objects',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/chrstnkln/urine',
    packages=setuptools.find_packages(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    tests_require=tests_require,
    extras_require={
        'test': tests_require,
        'lint': lint_requires,
        'dev': dev_requires
    },
    python_requires='>=3.6',
)



import struct
import zlib

from urine import __protocol_version__
from urine.extensions import UrineExtension
from urine.types import data_types


class UrineDecoder(object):
    _id_type_dict = {i: x for i, x in enumerate(data_types)}

    def __init__(self):
        self.data = None
        self.off = 0

        self._type_function_dict = {x: getattr(self, 'decode_' + x) for x in data_types}
        self._hash_extension_dict = {}

    def extend(self, extension):
        """Adds an extension to the decoder.

        Args:
            extension (UrineExtension): Extension inheriting and implementing UrineExtension

        Raises:
            ValueError: Two different extensions generate the same CRC32 hash (very unlikely)
        """
        if issubclass(extension, UrineExtension):
            if extension in self._hash_extension_dict.values():
                return

            crc32 = zlib.crc32(extension.__name__.encode('utf-8'))

            if crc32 in self._hash_extension_dict:
                raise ValueError(
                    'Extension hash duplicate detected, change extension name {}'.format(crc32)
                )

            self._hash_extension_dict[crc32] = extension

    def decode(self, data):
        """Decodes an object that was encoded with urine.

        Args:
            data (bytes, bytearray): Data to be decoded

        Raises:
            NotImplementedError: Unimplemented protocol version

        Returns:
            object: Decoded object
        """
        self.data = data
        self.off = 0

        protocol_version = self._read_protocol_version()
        if protocol_version != __protocol_version__:
            raise ValueError('Invalid protocol version: {} (not supported)'.format(protocol_version))

        return self._decode()

    def decode_extension(self):
        crc32 = self.decode_uint32()
        data_size = self.decode_uint32()

        data = self.data[self.off:self.off + data_size]
        self.off += data_size

        extension = self._hash_extension_dict[crc32]

        return extension.decode(data)

    def decode_bool(self):
        val = struct.unpack_from('?', self.data, self.off)[0]
        self.off += 1
        return val

    def decode_int8(self):
        val = struct.unpack_from('b', self.data, self.off)[0]
        self.off += 1
        return val

    def decode_int16(self):
        val = struct.unpack_from('<h', self.data, self.off)[0]
        self.off += 2
        return val

    def decode_int32(self):
        val = struct.unpack_from('<i', self.data, self.off)[0]
        self.off += 4
        return val

    def decode_int64(self):
        val = struct.unpack_from('<q', self.data, self.off)[0]
        self.off += 8
        return val

    def decode_uint8(self):
        val = struct.unpack_from('B', self.data, self.off)[0]
        self.off += 1
        return val

    def decode_uint16(self):
        val = struct.unpack_from('<H', self.data, self.off)[0]
        self.off += 2
        return val

    def decode_uint32(self):
        val = struct.unpack_from('<I', self.data, self.off)[0]
        self.off += 4
        return val

    def decode_uint64(self):
        val = struct.unpack_from('<Q', self.data, self.off)[0]
        self.off += 8
        return val

    def decode_bignum(self):
        signed = self.decode_bool()
        parts_len = self.decode_uint32()
        parts = [self.decode_uint64() for _ in range(parts_len)]

        final_hex = '0x'
        for part in parts:
            final_hex += hex(part)[2:]

        val = int(final_hex, 16)
        return -val if signed else val

    def decode_float(self):
        val = struct.unpack_from('<d', self.data, self.off)[0]
        self.off += 8
        return val

    def decode_complex(self):
        real, imag = struct.unpack_from('<dd', self.data, self.off)
        self.off += 16
        return complex(real, imag)

    def decode_bytes(self):
        data_size = self.decode_uint32()
        data = self.data[self.off:self.off + data_size]
        self.off += data_size

        return bytes(data)

    def decode_bytearray(self):
        data_size = self.decode_uint32()
        data = self.data[self.off:self.off + data_size]
        self.off += data_size

        return data

    def decode_string(self):
        str_size = self.decode_uint32()
        string = self.data[self.off:self.off + str_size].decode('utf-8')
        self.off += str_size

        return string

    def decode_list(self):
        lst_len = self.decode_uint32()
        return [self._decode() for _ in range(lst_len)]

    def decode_tuple(self):
        tpl_len = self.decode_uint32()
        return tuple(self._decode() for _ in range(tpl_len))

    def decode_set(self):
        st_len = self.decode_uint32()
        return set([self._decode() for _ in range(st_len)])

    def decode_frozenset(self):
        st_len = self.decode_uint32()
        return frozenset((self._decode() for _ in range(st_len)))

    def decode_dict(self):
        dct_len = self.decode_uint32()
        dct = {}

        for _ in range(dct_len):
            key = self._decode()
            val = self._decode()
            dct[key] = val

        return dct

    def decode_range(self):
        start = self._decode()
        stop = self._decode()
        step = self._decode()

        return range(start, stop, step)

    def decode_class(self):
        name = self.decode_string()
        attrs_len = self.decode_uint32()
        attrs_dct = {}

        for _ in range(attrs_len):
            key = self.decode_string()
            val = self._decode()
            attrs_dct[key] = val

        return type(name, (object,), attrs_dct)()

    def decode_none(self):
        return None

    def _decode(self):
        type_str = self._read_type()
        return self._type_function_dict[type_str]()

    def _read_protocol_version(self):
        major = self.decode_uint8()
        minor = self.decode_uint8()

        return '.'.join((str(major), str(minor)))

    def _read_type(self):
        type_id = self.decode_uint8()
        return self._id_type_dict[type_id]



import urine


def test_range():
    values = [range(10), range(3, 5), range(20, 300, 2)]
    for val in values:
        enc = urine.encode(val)
        dec = urine.decode(enc)

        assert dec == val
        assert isinstance(dec, range)
